//-< BasicMetaobject.java >------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 18-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Basic method methods implementation
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary> <code>BasicMetaobject</code> implements all the transaction logic. 
    /// Specifically pre/postDaemon are implemented to call begin/end read/write Access
    /// and call commit + abort Transaction.
    /// Locking is not implemented here, but in the subclasses. The functions begin/end
    /// read/write Access serve as hooks to do so.
    /// 
    /// </summary>
    /// <author>  K.A. Knizhnik
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public abstract class BasicMetaobject:Metaobject
    {
        /// <summary> Invocation of <code>preDaemon</code>  is inserted by MOP geneator before each
        /// method invocation or object component access. See MetaObject 
        /// This "basic" implementation increases the accesscount of the object and if 
        /// the flag indicates change, calls begin read/write Access. 
        /// Also it load, or prepares the object by calling these functions on the 
        /// objects store
        /// 
        /// </summary>
        /// <param name="obj">an <code>ObjectHeader</code> value that is modified
        /// </param>
        /// <param name="attr">an <code>int</code> flag, saying what kind of modufucation
        /// </param>
        public override void  preDaemon(ObjectHeader obj, int attr)
        {
            CacheManager mng = CacheManager.GetCacheManager();
            mng.enter();
            lock(obj)
            {
                if (obj.accessCount == 0)
                {
                    lock(this)
                    {
                        obj.accessCount = 1;
                        //
                        // Setting of access count to non-zero prevents from
                        // object destruction by receiving deterioration 
                        // notification from server. Instead of this "invalid"
                        // object attribute will set to true.
                        //
                    }
                    lock(mng.globalCS)
                    {
                        mng.nestedTransactionCount += 1;
                        if (obj.opid != 0)
                        {
                            // persistent object
                            mng.removeFromCache(obj);
                        }
                    }
                    if (obj.opid != 0)
                    {
                        // presistent objects
                        if ((attr & Metaobject.MUTATOR) != 0)
                        {
                            beginWriteAccess(obj);
                        }
                        else
                        {
                            beginReadAccess(obj);
                        }
                        if (obj.invalidated)
                        {
                            //
                            // In case of using pessimistic scheme object is 
                            // already locked at this moment, so it can not be 
                            // more changed by another application before the end 
                            // of transaction. 
                            // And with optimistic scheme... hmmm. it nice to be 
                            // optimist.
                            //
                            lock(this)
                            {
                                obj.invalidated = false;
                                destroyObject(obj);
                            }
                        }
                        if ((obj.state & ObjectState.DESTRUCTED) != 0)
                        {
                            obj.storage.load(obj.opid, obj);
                        }
                        else if ((obj.state & ObjectState.RAW) != 0)
                        {
                            obj.storage.prepare(obj);
                        }
                    }
                }
                else
                {
                    obj.accessCount += 1;
                }
            }
        }
		
        /// <summary> Invocation of <code>postDaemon</code>  is inserted by MOP geneator after each
        /// method invocation or object component access. See also MetaObject 
        /// This "basic" implementation decreases the accesscount of the object and 
        /// if that hits 0 calls endAccess(obj). In that case, the object is removed from 
        /// the  cache  manager and detroyObject is called. If that was the end of
        /// nested Trasaction, either commitTransaction, or abortTransaction, 
        /// depending on if the flag indicates Exception is called.
        ///
        /// </summary>
        /// <param name="obj">an <code>ObjectHeader</code> of object that has been changed
        /// </param>
        /// <param name="attr">an <code>int</code>, indicating where/how the chage happened. An OR
        /// combination of the static flags below
        /// </param>
        public override void  postDaemon(ObjectHeader obj, int attr)
        {
            CacheManager mng = CacheManager.GetCacheManager();
            lock(obj)
            {
                if (obj.accessCount == 1)
                {
                    if (obj.opid != 0)
                    {
                        endAccess(obj);
                    }
                    lock(mng.globalCS)
                    {
                        if (obj.opid != 0)
                        {
                            // persistent object
                            if ((obj.state & (ObjectState.NOTIFY | ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) == 0)
                            {
                                mng.insertInCache(obj);
                            }
                        }
                        if (--mng.nestedTransactionCount == 0)
                        {
                            if ((attr & Metaobject.EXCEPTION) != 0)
                            {
                                abortTransaction();
                            }
                            else
                            {
                                commitTransaction();
                            }
                        }
                    }
                    lock(this)
                    {
                        //
                        // We delay decrement of access counter till this moment
                        // to prevent object destruction on receiving deterioration
                        // notification from server. Synchronized construction 
                        // for metaobject object provides mutual exclusion with 
                        // objectInvalidate() method.
                        //
                        obj.accessCount = 0;
                        if (obj.invalidated)
                        {
                            obj.invalidated = false;
                            lock(mng)
                            {
                                mng.removeFromCache(obj);
                                destroyObject(obj);
                            }
                        }
                    }
                }
                else
                {
                    obj.accessCount -= 1;
                }
            }
            mng.leave();
        }
		
        /// <summary> Assign persistent identifier to the transient object of persistent
        /// capable class. The object is placed in the same storage as 
        /// persistent object containing reference to this object unless
        /// explicit storage attachment was specified for this object.
        /// 
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value that is made persistent
        /// </param>
        /// <param name="parent">a <code>ObjectHeader</code> into whose storage obj is placed, if 
        /// obj is new
        /// 
        /// </param>
        protected internal override void  makePersistent(ObjectHeader obj, ObjectHeader parent)
        {
            if (obj.storage == null)
            {
                //
                // By default transient object is placed in the same storage as 
                // persitent object referenced this object.
                //
                obj.storage = parent.storage;
            }
            else
            {
                // 
                // Interdatabase references are prohibited
                //
                System.Diagnostics.Debug.Assert(obj.storage.database == parent.storage.database);
            }
            obj.storage.allocate(obj, 0);
            //
            // Object is linked in transaction list after persistent object, 
            // from which it is referenced, so this object will handled by
            // commitTransaction() method soon after it complets with parent 
            // object.
            //
            CacheManager mng = CacheManager.GetCacheManager();
            Goods.CacheManager.linkAfter(parent, obj);
            obj.state |= ObjectState.TRANSWRITE | ObjectState.DIRTY;
        }
		
        /// 
        /// <summary> Object thrown by GC from weak cache are linked into forgottenObjectList
        /// list and notifications are send to servers at the transaction
        /// commit time. This is done because clearing weak reference by GC 
        /// is not synchronized with invocation of finalize() method. So there is
        /// no warrenty that in the moment between object is removed from weak cache
        /// and ionvocation of finilize method, the object with the same identifier
        /// will not be retirieved from the server. Delayed strategy of sending
        /// forget message together with checking object for presence in the cache 
        /// before sending foget message to the server makes sending of wrong
        /// forgot message not possible. 
        /// OId's of Object that were "forgotten" are collected and handed to 
        /// Storage.forgetObjects
        ///
        /// </summary>
        /// <param name="mng">a <code>CacheManager</code> value
        /// 
        /// </param>
        protected internal static void  sendForgottenObjectsToServers(CacheManager mng)
        {
            ForgottenObject list;
            lock(mng)
            {
                list = mng.forgottenObjectList;
                mng.forgottenObjectList = null;
            }
            while (list != null)
            {
                Storage store = list.storage;
                int n = 0;
                for (ForgottenObject obj = list, prev = null; obj != null; obj = obj.next)
                {
                    if (obj.storage == store)
                    {
                        Object o = store.objectCache.get(obj.opid);
                        if (o == obj)
                        {
                            store.objectCache.remove(obj.opid);
                            n += 1;
                        }
                        else if (o == null)
                        {
                            n += 1;
                        }
                        else
                        {
                            if (prev == null)
                            {
                                list = obj.next;
                            }
                            else
                            {
                                prev.next = obj.next;
                            }
                            continue;
                        }
                    }
                    prev = obj;
                }
                if (n > 0)
                {
                    int[] opids = new int[n];
                    for (ForgottenObject obj = list, prev = null; obj != null; obj = obj.next)
                    {
                        if (obj.storage == store)
                        {
                            opids[--n] = obj.opid;
                            if (prev == null)
                            {
                                list = obj.next;
                            }
                            else
                            {
                                prev.next = obj.next;
                            }
                            continue;
                        }
                        prev = obj;
                    }
                    store.forgetObjects(opids);
                }
            }
        }
		
        /// <summary> At this moment there are no more active methods for 
        /// classes controlled by this metaobject (or metaobjects derived from it).
        /// This method will try to commit all transactions in all opened databases.
        /// Until this method is completed no other invocation or access to
        /// objects controled by this MOP is possible (it is forced by declaring
        /// this static method "synchronized" as well as static method fixObject(),
        /// which is called before any access to the object). 
        ///
        /// First sendForgottenObject is called. Then, for all modifed objects (all
        /// others are release by metaobject.releaseObject()) we find it's database
        /// and collect all object involved in the trasaction. If any is invalidated, 
        /// the transaction is aborted. 
        ///
        /// Each objects storage store(obj) function is called, which packs the byte
        /// into the tranaction buffer. Then the storage's commitTransaction is called,
        /// and if the trasaction spans several storages, those storages 
        /// commitSubtransaction are called. This send the bytes to the server, and if
        /// anything goes wrong, abortTransaction is called. 
        ///
        /// If all went well, obj.metaObject.commitchanegs is called for every object
        /// </summary>
        protected internal static void  commitTransaction()
        {
            ObjectHeader next;
            CacheManager mng = CacheManager.GetCacheManager();
            sendForgottenObjectsToServers(mng);
            ObjectHeader transObjects = mng.transObjects;
            while (transObjects.next != transObjects)
            {
                ObjectHeader obj = transObjects.next;
                if ((obj.state & ObjectState.DIRTY) == 0)
                {
                    // 
                    // As far as all modified objects are placed at the head of 
                    // the list, and we are navigating through the list startig 
                    // from the head, finding of non-modified object means that
                    // there are no more other modified objects. So no more 
                    // other transactions has to be send to the server. 
                    // We only should release all locks hold by these transactions.
                    //
                    do 
                    {
                        next = obj.next;
                        obj.metaobject.releaseObject(obj);
                    }
                    while ((obj = next) != transObjects);
                    break;
                }
                int nTransServers = 0;
                Storage coordinator = obj.storage;
                Database db = coordinator.database;
                Storage[] servers = new Storage[db.storages.Length];
				
                do 
                {
                    // find all tranasction objects from the same database
                    Storage storage = obj.storage;
                    if (storage.database == db)
                    {
                        if (obj.invalidated)
                        {
                            abortTransaction(mng, db);
                            goto loop;
                        }
                        if (!storage.isInvolvedInTransaction())
                        {
                            if (storage.id < coordinator.id)
                            {
                                coordinator = storage;
                            }
                            servers[nTransServers++] = storage;
                        }
                        storage.store(obj);
                    }
                    obj = obj.next;
                }
                while (obj != transObjects);
				
                if (nTransServers != 0)
                {
                    //
                    // commit local transaction or coordinator part of global one
                    //
                    int tid = coordinator.commitTransaction(nTransServers, servers);
                    if (tid == - 1)
                    {
                        abortTransaction(mng, db);
                        continue;
                    }
                    if (nTransServers > 1)
                    {
                        // global transaction
                        for (int i = 0; i < nTransServers; i++)
                        {
                            if (servers[i] != coordinator)
                            {
                                servers[i].commitSubtransaction(coordinator, nTransServers, servers, tid);
                            }
                        }
                        for (int i = 0; i < nTransServers; i++)
                        {
                            if (!servers[i].waitTransactionCompletion())
                            {
                                abortTransaction(mng, db);
                                goto loop;
                            }
                        }
                    }
                }
                //
                // Transaction was successfully commited
                //
                for (obj = transObjects.next; obj != transObjects; obj = next)
                {
                    next = obj.next;
                    if (obj.storage.database == db)
                    {
                        obj.metaobject.commitObjectChanges(obj);
                    }
                }
            loop: ;
            }
        }
		
        /// <summary>  <code>abortTransaction</code> obviously aborts the Transaction in progress.
        /// This is done by calling undoObjectChanges on each Objects metaobject, and 
        /// calling abortTransaction on the DB.
        ///
        /// </summary>
        /// <param name="mng">a <code>CacheManager</code> that provides the list of objects
        /// that participate in the Transaction. (mng.transObjects)
        /// </param>
        /// <param name="db">a <code>Database</code>, in which all participating Objects must be
        /// 
        /// </param>
        protected internal static void  abortTransaction(CacheManager mng, Database db)
        {
            ObjectHeader obj, next;
			
            ObjectHeader transObjects = mng.transObjects;
            for (obj = transObjects.next; obj != transObjects; obj = next)
            {
                next = obj.next;
                if (obj.storage.database == db)
                {
                    obj.metaobject.undoObjectChanges(obj);
                }
            }
            db.abortTransaction();
        }
		
        /// <summary>  <code>abortTransaction</code> aborts for the static 
        /// CacheManager.getCacheManger (per thread) and the database that the objects
        /// in that are in. calls 
        ///
        /// </summary>
        protected internal static void  abortTransaction()
        {
            CacheManager mng = CacheManager.GetCacheManager();
            ObjectHeader transObjects = mng.transObjects;
            while (transObjects.next != transObjects)
            {
                abortTransaction(mng, transObjects.next.storage.database);
            }
        }
		
        /// <summary>  <code>commitObjectChanges</code> is called during commitTransaction. The 
        /// object is removed from the transaction list of the cache manager. If the obj
        /// is invalidated, it's destroyed (this.destroyObject()).
        ///
        /// </summary>
        /// <param name="obj">a <code>Persistent</code> object to commit
        /// 
        /// </param>
        protected internal override void  commitObjectChanges(ObjectHeader obj)
        {
            lock(this)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                System.Diagnostics.Debug.Assert((obj.state & (ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) != 0);
                CacheManager.unlink(obj); // remove object from transaction list
                obj.state &= ~ (ObjectState.TRANSREAD | ObjectState.TRANSWRITE | ObjectState.XLOCKED | ObjectState.SLOCKED | ObjectState.DIRTY | ObjectState.NEW);
                if (obj.invalidated)
                {
                    obj.invalidated = false;
                    destroyObject(obj);
                }
                else if ((obj.state & ObjectState.NOTIFY) == 0)
                {
                    mng.insertInCache(obj);
                }
            }
        }
		
        /// <summary> <code>undoObjectChanges</code> is called during an abortTransaction. 
        /// The object is removed  from the cache managers Trasaction list and 
        /// if locked, unlocked. If the object has been modified (dirty), new objects 
        /// are destroyed and old removed from their storage (storage.forgetObject())
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        protected internal override void  undoObjectChanges(ObjectHeader obj)
        {
            lock(this)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                System.Diagnostics.Debug.Assert((obj.state & (ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) != 0);
                CacheManager.unlink(obj); // remove object from transaction list
				
                if ((obj.state & (ObjectState.SLOCKED | ObjectState.XLOCKED)) != 0)
                {
                    obj.storage.unlock(obj, Protocol.lck_none);
                }
				
                if ((obj.state & ObjectState.DIRTY) != 0)
                {
                    if ((obj.state & ObjectState.NEW) == 0)
                    {
                        //
                        // All persistent objects should be reloaded from the server.
                        //
                        destroyObject(obj);
                    }
                    else
                    {
                        obj.storage.forgetObject(obj);
                        obj.opid = 0; // object beacomes transsient once again
                        obj.storage = null;
                    }
                }
                else if ((obj.state & ObjectState.NOTIFY) == 0)
                {
                    mng.insertInCache(obj);
                }
                obj.state &= ~ (ObjectState.TRANSREAD | ObjectState.TRANSWRITE | ObjectState.XLOCKED | ObjectState.SLOCKED | ObjectState.DIRTY | ObjectState.NEW);
            }
        }
		
        /// <summary>  <code>preloadObject</code> is called by Storage.read()  when it's bytes
        /// have been received and just before they are copied into it's data member 
        /// The function adjusts the cache Managers cacheSize appropriately
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> that has been read (from the net)
        /// </param>
        /// <param name="size">an <code>int</code>, how many bytes the data member will hold
        /// 
        /// </param>
        protected internal override void  preloadObject(ObjectHeader obj, int size)
        {
            CacheManager mng = CacheManager.GetCacheManager();
            lock(mng)
            {
                if (obj.data != null)
                {
                    mng.cacheSize0 -= obj.data.Length;
                }
                else
                {
                    mng.cacheSize0 += CacheManager.objectAllocationOverhead + obj.desc.fixedSize;
                }
                mng.cacheSize0 += size;
            }
        }
		
        /// <summary>  <code>loadObject</code> is not so much the loading of the Object, ,but a 
        /// notification thereof. The object is really loaded by the Storage class  
        /// (in prepare(obj) method) and after the deed is done, this method is called. 
        ///  This only increases the cache size appropriately
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object that has been loaded
        /// 
        /// </param>
        protected internal override void  loadObject(ObjectHeader obj)
        {
            CacheManager mng = CacheManager.GetCacheManager();
            lock(mng)
            {
                mng.cacheSize0 -= CacheManager.objectAllocationOverhead + obj.data.Length + obj.desc.fixedSize;
            }
        }
		
        /// <summary> <code>forgetObject</code> is called when the objects finalize() is called, if 
        /// it's opid is still valid. If all data were local, this would mean the object
        /// could be desroyed because it's not reachable anymore. But because more 
        /// references may exist at the servers end, this basically reduces the reference 
        /// count at the servers side. 
        /// Server side garbage collection will eventually reclaim the object space, if
        /// it really has become unreachable. 
        /// The function adjusts the cache manager's cache size and links the object into the
        /// forgottenObjectList of the cache manager
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> that has become localy unreachable
        /// 
        /// </param>
        protected internal override void  forgetObject(ObjectHeader obj)
        {
            CacheManager cm = obj.storage.database.cacheManager;
            if (cm == null)
            {
                cm = CacheManager.defaultCacheManager;
            }
            cm.forgetObject(obj);
        }
		
		
        /// <summary> <code>releaseObject</code> is called during a commitTransaction, for objects
        /// that have not been modified. The Object is removed from the Transaction list
        /// and unlocked (if it was). If NOTIFY is set on the object we keep it in cache.
        ///  DIRTY,NEW,TANS*,*LOCK flags are cleared.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> to be released.
        /// 
        /// </param>
        protected internal override void  releaseObject(ObjectHeader obj)
        {
            lock(this)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                System.Diagnostics.Debug.Assert((obj.state & (ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) != 0);
                CacheManager.unlink(obj); // remove object from transaction list
				
                if ((obj.state & (ObjectState.SLOCKED | ObjectState.XLOCKED)) != 0)
                {
                    obj.storage.unlock(obj, Protocol.lck_none);
                }
                if ((obj.state & ObjectState.NOTIFY) == 0)
                {
                    mng.insertInCache(obj);
                }
                obj.state &= ~ (ObjectState.TRANSREAD | ObjectState.TRANSWRITE | ObjectState.XLOCKED | ObjectState.SLOCKED | ObjectState.DIRTY | ObjectState.NEW);
            }
        }
		
        /// <summary> Destoroy all object references to make it possible to provide
        /// more unaccessible objects to GC 
        /// The objects state is set to DESTRUCTED | RAW
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        protected internal override void  destroyObject(ObjectHeader obj)
        {
            if ((obj.state & ObjectState.DESTRUCTED) == 0)
            {
                obj.reset();
                obj.state |= ObjectState.DESTRUCTED | ObjectState.RAW;
                obj.storage.throwObject(obj);
            }
        }
		
        /// <summary> This method is asynchronously called by server storag agent when
        /// deterioration notification from server is received. 
        /// If the object does not participate in a current Transaction, it is removed 
        /// from the cache and detroyed. Otherwise the invalidated flag is set and 
        /// detruction happends after commit. 
        /// If NOTIFY is set on the object, the associated CondEvent's signal is called
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to be invalidated.
        /// 
        /// </param>
        protected internal override void  invalidateObject(ObjectHeader obj)
        {
            lock(this)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                if ((obj.state & ObjectState.DESTRUCTED) == 0)
                {
                    if (obj.accessCount != 0 || (obj.state & (ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) != 0)
                    {
                        //
                        // It is not possible to detroy object immediatly, 
                        // just mark it as invalid and delay destruction till end
                        // of access to the object or end of transaction.
                        //
                        obj.invalidated = true;
                    }
                    else
                    {
                        obj.invalidated = false;
                        lock(mng)
                        {
                            mng.removeFromCache(obj);
                            destroyObject(obj);
                        }
                    }
                }
                if ((obj.state & ObjectState.NOTIFY) != 0)
                {
                    CondEvent event_Renamed = (CondEvent) obj.storage.database.notifications.get(obj);
                    if (event_Renamed != null)
                    {
                        event_Renamed.signal();
                    }
                }
            }
        }
		
        /// <summary> Enable or disable notification on receving invalidation messages from
        /// the server for this object. If enable, set NOTIFY flag on the object and
        /// remove from cache (?). Else unset the flag and insert into cache, if not 
        /// accessed.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> to make the notification changes for
        /// </param>
        /// <param name="enabled">a <code>boolean</code>, whether notifications are used for this
        /// object
        /// 
        /// </param>
        protected internal override void  notifyOnModification(ObjectHeader obj, bool enabled)
        {
            lock(this)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                lock(mng)
                {
                    if (enabled)
                    {
                        obj.state |= ObjectState.NOTIFY;
                        if ((obj.state & ObjectState.CACHED) != 0)
                        {
                            mng.removeFromCache(obj);
                        }
                    }
                    else
                    {
                        obj.state &= ~ ObjectState.NOTIFY;
                        if (obj.accessCount == 0 && (obj.state & (ObjectState.CACHED | ObjectState.TRANSREAD | ObjectState.TRANSWRITE)) == 0)
                        {
                            mng.insertInCache(obj);
                        }
                    }
                }
            }
        }
		
		
		
        /// <summary> By incrementing nested transaction count we delay implicit transaction 
        /// commit until the moment when counter is explicitly decremented by
        /// endNestedTransactionMethod()
        ///
        /// </summary>
        public static void  beginNestedTransaction()
        {
            CacheManager mng = (CacheManager) CacheManager.GetCacheManager();
            lock(mng.globalCS)
            {
                mng.nestedTransactionCount += 1;
            }
        }
		
        /// <summary> <code>endNestedTransaction</code> decreases the cache managers 
        /// nestedTransaction counter, and if that hits 0 calls commitTransaction 
        ///
        /// </summary>
        public static void  endNestedTransaction()
        {
            CacheManager mng = (CacheManager) CacheManager.GetCacheManager();
            lock(mng.globalCS)
            {
                if (--mng.nestedTransactionCount == 0)
                {
                    commitTransaction();
                }
            }
        }
    }
}