//-< Metaobject.java >-----------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  3-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Abstraction of metaobject protocol used by JAVAMOP preprocessor
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary> <code>Metaobject</code> is a base class for all Transaction + locking strategies.
    /// 
    /// ObjectHeader holds a MetaObject, which is called when the object is
    /// modified or read. This also means you can have different strategies for different
    /// classes, or even object if you wish. The default is PessimisticRepeatableRead.
    /// This class has abstract functions to handle loading and
    /// unloading of Object over the wire, and to handle Transactions. Specifically
    /// their failure strategies are implemented by different implementations of the 
    /// various derived classes.
    /// Most functions are implemented in the BasicMetaObject, but the begin/end * Access
    /// functions are implemented in the different derived classes.
    /// JAVAMOP, the MetaObjectProtocol postprocessor inserts all necessary code
    /// to call MetaObject functions into the compiled byte code.
    /// </summary>
    /// <author> K.A. Knizhnik
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public abstract class Metaobject
    {
        /// <summary> Invocation of this <code>preDaemon</code> is inserted by MOP generator 
        /// before each method invocation or object component access. (Other database
        /// implementation call this method fetch(), to indicate that this causes
        /// the data of the Object to be fetched, if it's not already there) 
        /// The exact call that is inserted is  super.metaobject.preDaemon(this, attr);
        ///
        /// </summary>
        /// <param name="obj">The <code>ObjectHeader</code> of loaded object.
        /// </param>
        /// <param name="attr">an <code>int</code> value, indicating the kind of access. An OR 
        /// combination of the static flags below.
        /// 
        /// </param>
        abstract public void  preDaemon(ObjectHeader obj, int attr);
		
        /// <summary> Invocation of this <code>postDaemon</code> is inserted by MOP generator after 
        /// each  method invocation or object component access.
        /// The inserted code is something like:  super.metaobject.postDaemon(this,attr,flag)
        ///
        /// </summary>
        /// <param name="obj">an <code>ObjectHeader</code> of object hat has been changed
        /// </param>
        /// <param name="attr">an <code>int</code>, indicating where/how the change happened. An OR
        /// combination of the static flags below
        /// </param>
        abstract public void  postDaemon(ObjectHeader obj, int attr);
		
        /// <summary>  <code>MUTATOR</code> == 1 , means object can be changed
        /// </summary>
        public const int MUTATOR = 1;
		
        /// <summary>  <code>VARIABLE</code> == 2 , means access to non-self instance variable 
        /// </summary>
        public const int VARIABLE = 2;
		
        /// <summary>  <code>CONSTRUCTOR</code> == 4 means  wrapped method is constructor
        /// </summary>
        public const int CONSTRUCTOR = 4;
		
        /// <summary>  <code>EXCEPTION</code> == 8, means method is terminated by exception 
        /// </summary>
        public const int EXCEPTION = 8;
		
				
        /// <summary>  <code>preloadObject</code> is called by Storage.read()  when it's bytes
        /// have been received and just before they are copied into it's data member 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> that has been read (from the net)
        /// </param>
        /// <param name="size">an <code>int</code>, how many bytes the data member will hold
        /// 
        /// </param>
        abstract protected internal void  preloadObject(ObjectHeader obj, int size);
		
        /// <summary>  <code>loadObject</code> is not so much the loading of the Object, ,but a 
        /// notification thereof. The object is really loaded by the Storage class  
        /// (in prepare(obj) method) and after the deed is done, this method is called. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object that has been loaded
        /// 
        /// </param>
        abstract protected internal void  loadObject(ObjectHeader obj);
		
        /// <summary> <code>forgetObject</code> is called when the objects finalize() is called, if 
        /// it's opid is still valid. If all data were local, this would mean the object
        /// could be destroyed because it's not reachable anymore. But because more 
        /// references may exist at the servers end, this basically reduces the reference 
        /// count at the servers side. 
        /// Server side garbage collection will eventually reclaim the object space, if
        /// it really has become unreachable. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> that has become locally unreachable
        /// 
        /// </param>
        abstract protected internal void  forgetObject(ObjectHeader obj);
		
        /// <summary> <code>releaseObject</code> is called during a commitTransaction, for objects
        /// that have not been modified.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> to be released.
        /// 
        /// </param>
        abstract protected internal void  releaseObject(ObjectHeader obj);
		
        /// <summary>  <code>commitObjectChanges</code> is called during commitTransaction. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to commit
        /// 
        /// </param>
        abstract protected internal void  commitObjectChanges(ObjectHeader obj);
		
        /// <summary> <code>undoObjectChanges</code> is called during an abortTransaction. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        abstract protected internal void  undoObjectChanges(ObjectHeader obj);
		
        /// <summary> <code>makeObjectHeader</code> should do just that, make the obj ObjectHeader.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to make ObjectHeader (or prepare for
        /// that)
        /// </param>
        /// <param name="parent">a <code>ObjectHeader</code> object "close" to the obj. May be 
        /// used , as by BasicMetaObject, to place the obj in the same storage as parent.
        /// 
        /// </param>
        abstract protected internal void  makePersistent(ObjectHeader obj, ObjectHeader parent);
		
        /// <summary>  <code>beginReadAccess</code> is called when the read Access begins. It is called
        /// from the (auto generated) preDeamon (in BasicMataObject) if the Mutator flag is 
        /// set. 
        /// Different implementations of this function can be found in the subclasses and 
        /// by eg locking or not locking the Object provide different strategies.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object that is about to be read.
        /// 
        /// </param>
        abstract protected internal void  beginReadAccess(ObjectHeader obj);
		
        /// <summary> <code>beginWriteAccess</code> is called when the write Access begins. It is 
        /// called from the (autogenerated) preDeamon (in BasicMataObject) if the Mutator
        /// flag is not set. 
        /// Different implementations of this function can be found in the subclasses and 
        /// by eg locking or not locking the Object provide different strategies.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object that is about to be read.
        /// 
        /// </param>
        abstract protected internal void  beginWriteAccess(ObjectHeader obj);
		
        /// <summary>  <code>endAccess</code> is called when the (current) access/function has ended, 
        /// and the object is not accessed anymore (the accesscount hits 0 ).
        /// It is called from the (autogenerated) postDeamon in BasiMetaObject.
        ///
        /// Different implementations of this function can be found in the subclasses and 
        /// by eg locking or not locking the Object provide different strategies.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object that is about to be read.
        /// 
        /// </param>
        abstract protected internal void  endAccess(ObjectHeader obj);
		
        /// <summary> This method is asynchronously called by server storag agent when
        /// deterioration notification from server is received. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to be invalidated.
        /// 
        /// </param>
        abstract protected internal void  invalidateObject(ObjectHeader obj);
		
        /// <summary> Destoroy all object references to make it possible to provide
        /// more unaccessible objects to GC 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        abstract protected internal void  destroyObject(ObjectHeader obj);
		
        /// <summary> Enable or disable notification on receving invalidation messages from
        /// the server for this object. 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> to make the notification changes for
        /// </param>
        /// <param name="status">a <code>boolean</code>, whether notifications are used for this
        /// object
        /// 
        /// </param>
        abstract protected internal void  notifyOnModification(ObjectHeader obj, bool status);
    }
}