//-< BULKLOAD.CXX >--------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     24-Apr-2007  K.A. Knizhnik  * / [] \ *
//                          Last update: 24-Apr-2007  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test for bulk set members loading 
//-------------------------------------------------------------------*--------*

#include "goods.h"
#include "dbscls.h"
#include <stdio.h>
#include <time.h>

USE_GOODS_NAMESPACE

#define N_RECORDS    100000
#define N_ITERATIONS 10
#define DATA_LEN     1024

class record : public object { 
  public: 
    ref<set_member> key;
    int             id;
    char            data[DATA_LEN];

    static int n_destructor_invocations;

    METACLASS_DECLARATIONS(record, object);

    record(int rid) : object(self_class) 
    { 
        char buf[64];
        sprintf(buf, "%05d", rid);
        id = rid;
        key = set_member::create(this, buf);
    }

    ~record() { 
        n_destructor_invocations += 1;
    }
};

class container : public B_tree { 
  public: 
    void insert(int rid) { 
        ref<record> rec = new record(rid);
        B_tree::insert(rec->key);
    }

    void initialize() const { 
        if (is_abstract_root()) { 
            ref<container> root = this;
            modify(root)->become(NEW container());
        } 
    }

    METACLASS_DECLARATIONS(container, B_tree);

    container() : B_tree(self_class) {}
};


field_descriptor& record::describe_components()
{
    return FIELD(key), FIELD(id);
}

field_descriptor& container::describe_components()
{
    return NO_FIELDS;
}

REGISTER(container, B_tree, pessimistic_repeatable_read_scheme);
REGISTER(record, object, pessimistic_repeatable_read_scheme);

int record::n_destructor_invocations;

int main(int argc, char* argv[]) 
{
    int i;
    database db;
    task::initialize(task::normal_stack);
    boolean bulk_load = False;

    for (i = 1; i < argc; i++) { 
        if (strcmp(argv[i], "-bulk") == 0) { 
            bulk_load = True;
        } else if (strcmp(argv[i], "-cache") == 0 && i+1 < argc) { 
            size_t cache_size = atoi(argv[++i])*1024*1024;
            pessimistic_repeatable_read_scheme.set_cache_limit(cache_size/2, cache_size/2);
        } else { 
            fprintf(stderr, "Usage: bulkload [-bulk] [-cache Mb]\n");
            return 1;
        }
    }
    if (db.open("bulkload.cfg")) { 
        ref<container> root;
        db.get_root(root);
        root->initialize();

        if (root->n_members == 0) { 
            time_t start = time(NULL);
            for (i = 0; i < N_RECORDS; i++) { 
                modify(root)->insert(N_RECORDS - i - 1);
            }
            printf("Elapsed time for inserting %d records: %d seconds\n", N_RECORDS, (int)(time(NULL) - start));
        }
        if (bulk_load) { 
            root->preload_members();
        }

        for (i = 0; i < N_ITERATIONS; i++) {
            time_t start = time(NULL); 
            int j = 0;
            for (ref<set_member> mbr = root->first; mbr != NULL; mbr = mbr->next) { 
                ref<record> rec = mbr->obj;
                assert(rec->id == j++);
            }
            assert(j == N_RECORDS);
            printf("Iteration %d: %d seconds\n", i, (int)(time(NULL) - start));        
        }
        db.close();
        printf("Number of destructor invocations=%d\n", record::n_destructor_invocations);
        return EXIT_SUCCESS;
    } else { 
        console::output("Failed to open database\n");
        return EXIT_FAILURE;
    }
}
