//-< CacheManager.java >---------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      4-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update:  4-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Object cache manager
//-------------------------------------------------------------------*--------*

package goodsjpi;

import java.util.Hashtable;

/**
 * <code>CacheManager</code> may be used on a per Thread basis (see attatch 
 * and detatch), or on a global basis (defaultCacheManager), to cache the
 * objects on client side and to keep transaction list.<br>
 * Two caches are kept, a last recently used and a frequesntly used. Cache
 * and transaction list are kept as double linked lists.
 *
 * @author  K.A. Knizhnik
 * @version 1.0
 */
public class CacheManager { 
  /**
   * Critical section to synchronize access to static variables of metaobject
   */
    Object globalCS = new Object();
  /**
   * Header of l2-list of objects involved in trabsaction. Modified objects 
   * are placed at the beginning of the list (transObjects.next...), while
   * not-modified objects involved in transaction are placed at the end of
   * the list (transObjects.prev...)
   * This is done to optimize sending of transaction to the storage server
   * (no extra sort or repacking of objects is needed).
   *
   * If object was first included in transaction list as "non-modified" and
   * later was modified, then it will be relinked to the beginning of the 
   * transaction list. 
   *
   * Object are kept in the transaction list until the end of transaction to 
   * prevent GC from deallocating them before end of tranasction.
   */
    Persistent transObjects;
  /**
   * Headers for two l2-lists used to implement enhanced LRU discipline
   * for replacing objects in object cache. Head of the list contains
   * most recently used object and tail - least recently used objects. 
   * To prevent object cache from replacing all object as a result of
   * database serach through large number of obejct (most of which will
   * not be used in future), standard LRU scheme was extended to more fit
   * specific of database applications. Object cache is divided into two 
   * part: primary cache and cache for frequently used objects. First time 
   * object is accessed, it is placed in primary cache. Object will be moved
   * to "frequently used objects" cache only when it is accessed more than
   * twice and it is not at the head of LRU list (case of several sucessive 
   * accesses to the object). Both cache parts are managed separately: when
   * total size of objects in the one part exceeds some limit value, least
   * recently used objects from this part are thrown away from clients memory
   * not affecting objects in another part of the cache.
   */
    Persistent lru0;
  /**
   * Frequesntly accessed object list (see above )
   **/
    Persistent lru1;
  /**
   * Total sizes of objects in the part of the cache (see explanation above)
   * Stub objects are not placed in the cache so thier size in not included
   * into this values. The total size occpied by persistent objects in 
   * client memory therefore can be greater than cacheSize0+cacheSize1. 
   */
    int cacheSize0;
  /**
   * size of frequently accessed list (see above)
   **/
    int cacheSize1;
  /**
   * Limits for total size of objects in both parts of the cache (see
   * explanation above). If size of some object exceed this limit values, 
   * it will be placed in the cache in any case. 
   */
    int cacheSizeLimit0 = 2*1024*1024;
  /**
   * Limit for the frequently accessed Cache size 2Mb
   **/
    int cacheSizeLimit1 = 2*1024*1024;

  /**
   * Minimal size of available free memory after reaching which object cache 
   * is cleaned to prevent hanging of thrown objects in the memory as
   * a result of presence references to this objects.  (1M)
   */
    long freeMemoryWatermark = 1024*1024;

  /**
   * Counter of nested transactions (or nested invocations of methods of
   * classes derived from Persistent class). BasicMetaobject implements 
   * implicit scheme of detection transaction boundaries. Transaction is 
   * committed when no more active methods exists for the classes controlled 
   * by this metaobject (or metaobjects derived from Basicmetaobject), 
   * i.e. when nestedTransactionCount is equal to zero. It is possbile
   * to manualy increment and decrement this count, so explictly
   * set transaction boundaries (see methods beginNestedTransaction() and
   * endNestedTransaction() below.
   */
    int nestedTransactionCount;

  /**
   * Head of a L1 list to save "forgotten objects". Forgotten 
   * means they were garbage collected, they are added here in 
   * Persistent.finalize(). Their Id's are sent back to the server, as 
   * finalizing is almost like locally deleting the object. This helps the 
   * server side garbage collection.
   */
    ForgottenObject forgottenObjectList;

  /**
   * Mutex used to provide exclusive access to CacheManager. By means of this mutex 
   * per-threads transactions sharing the same database connection are implemented. 
   * This mutex is used by BasicMetaobject when transaction model is PER_THREAD
   */
    Mutex transactionMutex = new Mutex();

  /**
   * Isolation level: Database.PER_THREAD_TRANSACTION or Database.PER_PROCESS_TRANSACTION
   */
    int isolationLevel = Database.PER_PROCESS_TRANSACTION;

  /**
   * Initialize all the double linked lists with empty Persistent objects.
   * The test whether a list is empty is thus list.next==list.prev .
   */
    public CacheManager() { 
	lru0 = new Persistent((Metaobject)null);
	lru0.next = lru0.prev = lru0;
	lru1 = new Persistent((Metaobject)null);
	lru1.next = lru1.prev = lru1;
	transObjects = new Persistent((Metaobject)null);
	transObjects.next = transObjects.prev = transObjects;
    }

  /**
   * <code>defaultCacheManager</code> is the one used if the current thread,
   * requesting the CacheManager (CahceManager.getCacheManager()) isnot 
   * attatched. In other words this is used as a "global" CacheManager.
   */
    static CacheManager defaultCacheManager = new CacheManager();

  /**
   *  <code>nAttached</code> is the count of how many Threads have been 
   * attatched. If this reaches the maximum (hashCleanupPeriod = 64), the
   * attatched threads are checked (if they are still alive).
   */
    static volatile int nAttached = 0; 

  /**
   *  <code>cacheManagerHash</code> holds the threads that have attatched
   * themselves. A thread may use new CahceManager().attatch() to do so. 
   * The keys are the threads, the values the managers
   */
    static java.util.Hashtable cacheManagerHash = new java.util.Hashtable();

  /**
   * <code>getCacheManager</code> the current manager. This is the per-thread
   * manager that has called attatch() or the default one if attatch has never
   * been called on the current thread.
   *
   * @return a <code>CacheManager</code> value, see above.
   */
    public static CacheManager getCacheManager() { 
	if (nAttached == 0) { 
	    return defaultCacheManager;
	}
	CacheManager mng; 
	Thread thread = Thread.currentThread();
	if (thread instanceof PThread) { 
	    mng = (CacheManager)((PThread)thread).getContext();
	} else { 
	    mng = (CacheManager)cacheManagerHash.get(thread);
	}
	if (mng == null) { 
	    mng = defaultCacheManager;
	}
	return mng;
    }
    
  /**
   * After this many attatch calls (64) the attatch method checks that all
   * attached threads (see cacheManagerHash) are still alive
   */
    static final int hashCleanupPeriod = 64;

  /** 
   * Associate cache manager with current thread. If the cleanupPeriod has 
   * been exceeded, check if all attatched threads are still alive.
   */
    public void attach() { 
	Thread thread = Thread.currentThread();
	if (thread instanceof PThread) { 
	    ((PThread)thread).setContext(this);
	    nAttached += 1;
	} else { 
	    if (++nAttached % hashCleanupPeriod == 0) { 
		java.util.Enumeration threads = cacheManagerHash.keys();
		while (threads.hasMoreElements()) { 
		    Thread thr = (Thread)threads.nextElement();
		    if (!thr.isAlive()) { 
			cacheManagerHash.remove(thr);
		    }
		}
	    }
	    cacheManagerHash.put(thread, this);
	}
    }

  /**
   * Detach cache manager from current thread 
   */
    public void detach() { 
	detach(Thread.currentThread());
    }

  /**
   * <code>detach</code> the given thread. Removes the associated cache 
   * manager from the hashtable (cacheManagerHash)
   *
   * @param thread a <code>Thread</code> to detatch
   */
    public void detach(Thread thread) {
	if (thread instanceof PThread) { 
	    ((PThread)thread).setContext(null);
	} else { 
	    cacheManagerHash.remove(thread);
	}
    }

  /**
   * Link the given object after the other. Ie, open the double linked list
   * after "after" and insert obj into it, so that after.next = obj and 
   * the previous after.next is obj.next
   * @param after a <code>Persistent</code> value, to link obj after
   * @param obj a <code>Persistent</code> value, to link in after "after"
   */
    public static void linkAfter(Persistent after, Persistent obj) { 
	(obj.prev = after).next = (obj.next = after.next).prev = obj;
    }

  /**
   *  <code>linkBefore</code> links "obj" before "before" such that obj.next
   *  == before.
   *
   * @param before a <code>Persistent</code> value to link object before
   * @param obj a <code>Persistent</code> value to link in.
   */
    public static void linkBefore(Persistent before, Persistent obj) { 
	(obj.next = before).prev = (obj.prev = before.prev).next = obj;
    }

  /**
   *  <code>unlink</code> the given object so that obj.next.prev = obj.prev
   * and obj.prev.next = obj.next 
   *
   * @param obj a <code>Persistent</code> value to unlink. 
   */
    public static void unlink(Persistent obj) { 
	Assert.that(obj.next.prev == obj && obj.prev.next == obj);
	obj.next.prev = obj.prev;
	obj.prev.next = obj.next;
	obj.next = obj.prev = null;
    }

  /**
   * Size of Persitent head plus size used by finilizer and object hash 
   * entry  (64)
   */
    static final int objectAllocationOverhead = 64;

  /**
   * Exclude object from LRU list so making it not available for 
   * cached objects LRU relpacing algorithm. This method is called
   * in two situations: 
   *  1) when object is accessed to prevent it from throwing away until
   *     it is accessed by active method.
   *  2) when modification notification for this object was received from
   *     server and instance of deteriorated object instance is removed from
   *     clients memory.
   * @param obj a <code>Persistent</code> value
   */
    public synchronized void removeFromCache(Persistent obj) { 
	if ((obj.state & Persistent.CACHED) != 0) { 
	    Assert.that((obj.state
			 & (Persistent.DESTRUCTED|Persistent.RAW)) == 0);
	    unlink(obj);
	    if ((obj.state & Persistent.USEFUL) == 0) { 
		if ((obj.state & Persistent.ACCESSED) != 0) { 
		    if (obj != lru0.next) { 
			obj.state |= Persistent.USEFUL;
		    }
		} else { 
		    obj.state |= Persistent.ACCESSED;
		}
		cacheSize0 -= objectAllocationOverhead + obj.desc.objectSize(obj);
	    } else { 
		cacheSize1 -= objectAllocationOverhead + obj.desc.objectSize(obj);
	    }
	    obj.state &= ~Persistent.CACHED;
	}
    }

    /**
     * Remove from cache all objects from the specified database
     * @param db database which objects should be removed from cache
     */
    public synchronized void removeFromCache(Database db) { 
	Persistent obj, next;
	for (obj = transObjects.next; obj != transObjects; obj = next) { 
	    next = obj.next;
	    if (obj.storage.database == db) { 
		unlink(obj);
		obj.opid = 0;
		
		if (obj.data != null) { 
		    cacheSize0 -= objectAllocationOverhead + obj.data.length + obj.desc.fixedSize;
		    obj.data = null;
		}	    
	    }
	}
	for (obj = lru0.next; obj != lru0; obj = next) { 
	    next = obj.next;
	    if (obj.storage.database == db) { 
		unlink(obj);
		cacheSize0 -= objectAllocationOverhead + obj.desc.objectSize(obj);
		obj.opid = 0;
		if (obj.data != null) { 
		    cacheSize0 -= objectAllocationOverhead + obj.data.length + obj.desc.fixedSize;
		    obj.data = null;
		}	    
	    }	
	}
	for (obj = lru1.next; obj != lru1; obj = next) { 
	    next = obj.next;
	    if (obj.storage.database == db) { 
		unlink(obj);
	    }	
	}
    }

    /**
     * Cache the static Runtime
     */
    static Runtime runtime = Runtime.getRuntime();

    /**
     * <code>insertInCache</code> insert object in one of two cache LRU lists
     * depending on its current
     * state. If total size of objects in this part of the cache exceeds limit 
     * value, least recently used objects will be removed from the cache
     * (and from client memory) until total size of objects becomes less or
     * equal than limit.
     *
     * @param obj a <code>Persistent</code> value
     */
    public synchronized void insertInCache(Persistent obj) { 
	Assert.that((obj.state 
		     & (Persistent.CACHED|Persistent.DESTRUCTED|Persistent.RAW)) == 0);
	long freeMemory = runtime.freeMemory();
	if ((obj.state & Persistent.USEFUL) == 0) { 
	    cacheSize0 += objectAllocationOverhead + obj.desc.objectSize(obj);
	    while ((freeMemory < freeMemoryWatermark
		    || cacheSize0 > cacheSizeLimit0) && lru0.prev != lru0) 
	    {
		Persistent victim = lru0.prev;
		cacheSize0 -= objectAllocationOverhead + victim.desc.objectSize(victim);
		victim.metaobject.destroyObject(victim);
		victim.state &= ~Persistent.CACHED;
		unlink(victim);
	    }
	    linkAfter(lru0, obj);
	} else { 
	    cacheSize1 += objectAllocationOverhead + obj.desc.objectSize(obj);
	    while ((freeMemory < freeMemoryWatermark 
		    || cacheSize1 > cacheSizeLimit1) && lru1.prev != lru1) 
	    {
		Persistent victim = lru1.prev;
		cacheSize1 -= objectAllocationOverhead + victim.desc.objectSize(victim);
		victim.metaobject.destroyObject(victim);
		victim.state &= ~Persistent.CACHED;
		unlink(victim);
	    }
	    linkAfter(lru1, obj);
	}
	obj.state |= Persistent.CACHED;
    }

  /**
   * Include object in transaction list. If object is modified it will be
   * placed at the head of the list, otherwise at the tail of the list.
   * If object already was included in transaction as "read-only", and 
   * later was modified, it will relinked to the head of the list.
   * @param obj a <code>Persistent</code> value
   */
    public void addToTransaction(Persistent obj) { 
	synchronized (transObjects) { 
	    if ((obj.state & Persistent.DIRTY) != 0) { 
		if ((obj.state & Persistent.TRANSWRITE) == 0) { 
		    if ((obj.state & Persistent.TRANSREAD) != 0) { 
			obj.state &= ~Persistent.TRANSREAD;
			unlink(obj);
		    }
		    linkAfter(transObjects, obj);
		    obj.state |= Persistent.TRANSWRITE;
		}
	    } else { 
		if ((obj.state & Persistent.TRANSREAD) == 0) { 
		    linkBefore(transObjects, obj);
		    obj.state |= Persistent.TRANSREAD;
		}
	    }
	}
    }

  /**
   * Set new watermark for minimal size of fee memory
   * @param watermark a <code>long</code> value
   */
    public void setFreeMemoryWatermark(long watermark) { 
	freeMemoryWatermark = watermark;
    }

  /**
   * Set new limits for object cache. See description of cache management
   * algorithm at the beggining of this file.
   * @param l0 an <code>int</code> value
   * @param l1 an <code>int</code> value
   */
    public void setCacheLimits(int l0, int l1) { 
	synchronized (globalCS) { 
	    cacheSizeLimit0 = l0;
	    cacheSizeLimit1 = l1;
	}
    }

  /**
   * This method is called by Metaobject.preDaemon to determine start of access to the database
   */
    public final void enter() {
	if (isolationLevel == Database.PER_THREAD_TRANSACTION) { 
	    transactionMutex.enter();
	}
    }

  /**
   * This method is called by Metaobject.postDaemon to determine end of access to the database
   */
    public final void leave() {
	if (isolationLevel == Database.PER_THREAD_TRANSACTION) { 
	    transactionMutex.leave();
	}
    }

    /**
     * Place object in forgotten list
     */ 
    public synchronized void forgetObject(Persistent obj) { 
	if (obj.data != null) { 
	    cacheSize0 -= objectAllocationOverhead + obj.data.length + obj.desc.fixedSize;
	    obj.data = null;
	}
	forgottenObjectList = new ForgottenObject(obj, forgottenObjectList);
    }
}
	

