package goodsjpi;

// Global imports
import java.util.Properties;
import java.io.IOException;
import java.io.InputStream;

/**
 * Class which defines the interface to a persistent store
 * @author Joel Crisp
 */
public abstract class PersistenceFactory
{
    /**
     * Root object of the object store
     */
    protected Object root;

    /**
     * Mapping from class and interface types to persistent
     * implementations of them
     * this contains lines like
     *    persistence.root=my.root.object.type
     *    persistence.java.util.Set=goodsjpi.PersistentHashSet
     *    persistence.java.util.List=goodsjpi.PersistentArrayList
     */
    protected static Properties persistenceMap;

    /**
     * Instance of the store specific implementation of
     * this factory
     */
    private static PersistenceFactory factory;

    static
    {
	persistenceMap=new Properties();
	
	InputStream source;
	ClassLoader loader=PersistenceFactory.class.getClassLoader();
	if(loader==null) source=ClassLoader.getSystemResourceAsStream("persistence.properties");
	else source=loader.getResourceAsStream("persistence.properties");
	try
	    {
		persistenceMap.load(source);
	    }
	catch(IOException ioe)
	    {
	    }
	finally
	    {
		try
		    {
			if(source!=null) source.close();
		    }
		catch(IOException ioe)
		    {
		    }
	    }
    }

    /**
     * Create the new persistence factory
     */
    protected PersistenceFactory() 
    {
    }

    /**
     * Get the instance of the persistence factory
     */
    public static final PersistenceFactory get()
    {
	if(factory==null) 
	    {
		try
		    {
			Class factoryClazz=getImplClass(PersistenceFactory.class);
			factory=(PersistenceFactory)factoryClazz.newInstance();
		    }
		catch(InstantiationException ie)
		    {
			System.err.println("FATAL: Can't instantiate persistence factory");
			System.exit(-1);
		    }
		catch(IllegalAccessException iae)
		    {
			System.err.println("FATAL: Can't instantiate persistence factory");
			System.exit(-2);
		    }
	    }
        return factory;
    }

    /**
     * Get an object instance which is compatible with the
     * required class and is also persistent
     * @param required the required class or interface type
     */
    public static final Object get(Class required)
    {
        return get().getImplementation(required);
    }

    /**
     * Get an object instance which is compatible with the
     * required class and is also persistent
     * @param required the required class or interface type
     * @param types array of types for the constructor
     * @param params array of parameters for the constructor
     */
    public static final Object get(Class required,Class types[],Object params[])
    {
        return get().getImplementation(required,types,params);
    }

    /**
     * Get the implementation type for the requested type
     * @param required a class or interface type for which
     *   a persistent implementation is required
     * @return the class type implementing persistence for the required type
     */
    protected static Class getImplClass(Class required)
    {
        String candidate=persistenceMap.getProperty("persistence."+required.getName());
        if(candidate==null) System.err.println("Can't map "+required);
	System.err.println("Mapping class "+required+" to "+candidate);
        Class implClass;
        ClassLoader loader=PersistenceFactory.class.getClassLoader();
        try
            {
                if(loader==null) implClass=Class.forName(candidate);
                else implClass=loader.loadClass(candidate);
            }
        catch(ClassNotFoundException cnfe)
            {
                cnfe.printStackTrace();
                return null;
            }
        return implClass;
    }

    /**
     * Handle on the datastore specific getImplementation method
     * @param required class for which a persistent object is required
     */
    protected abstract Object getImplementation(Class required);

    /**
     * Handle on the datastore specific getImplementation method
     * @param required class for which a persistent object is required
     * @param types array of types for the constuctor arguments
     * @param params array of parameters for the constructor
     */
    protected abstract Object getImplementation(Class required,Class types[],Object params[]);

    /**
     * Get the root object of the data store
     * @return an object of type mapped by the entry persistence.root in the config file
     */
    public static Object getRoot() throws IOException
    {
	return get().getRootImpl();
    }

    /**
     * Handle on the datastore specific implementation of the getRoot method
     * @return an object of type mapped by the entry persistence.root in the config file
     */
    protected abstract Object getRootImpl() throws IOException;
}
