package org.dcg.gen;

import java.util.Stack;

import org.dcg.datamgr.DBConnector;
import org.dcg.datamgr.databasejobs.DatabaseIndexJob;
import org.dcg.datamgr.databasejobs.DatabaseJob;
import org.dcg.gen.GeneralJob;

/**
 * @author dacosta
 * 
 * This is the generic ThreadPoolMgr that can either be used as a 'normal' PoolMgr or as a DatabaseConnectionMgr.</ br>
 * The latter implies that each thread has got an active handle to its database (any database, because this can be handled in the ThreadPoolWorker in conjunction with the DBConnector.</ br>
 * Basic structure:</ br>
 * 	There are two queues. One for the pending jobs that need to be executed and another that contains the inactive and not timed-out threads.</ br>
 * 	ANY job executed is taken of the top of the pending queue by a thread which is free to process it.</ br>
 * 	It is the thread itself that looks at the queue. Once obtained the queue is instantly updated.</ br>
 * 	Next the thread starts processing whatever job it holds.</ br>
 */
public abstract class ThreadPoolMgr {
	protected Stack _pendingJobs = new Stack(); // A queue of pending jobs
	protected Stack _availableThreads = new Stack(); // A queue of all available threads
	// ThreadGroup to control the threads within the PoolMgr
	protected ThreadGroup _totalThreadGroup = new ThreadGroup("genTG");

	// Required also for the ThreadPoolWorkers
	protected int _maxThreads = -1; // max number of threads that can be created
	protected int _minThreads = -1; // min number of threads that must always be present
	protected int _nrOfThreadInstances = 0; // Counter to keep track of the instances created

	protected int _maxIdleTime = -1; // max idle time after which a thread may be killed
	protected boolean _debug = false; // The debug flag
	protected DBConnector _databaseConnector = null;
	protected String _service;
	protected ThreadPoolMgrStats stats = new ThreadPoolMgrStats();

	/**
	 * Each thread in the pool is an instance of this class.<br>
	 * This thread-instance is actually 'the thing' that executes the job handed to it.<br>
	 * An example would be an employee in a company.<br>
	 * Basically this is the 'worker/ labour class' that is available to the Mgr/ the pool
	 */
	protected class PoolThread extends ThreadPoolWorker {
		
		/**
		 * Constructor for PoolThread as 'normal' PoolMgr.<br>
		 * @param tg is the ThreadGroup to which this thread belongs
		 * @param lock is the object that takes care of overall PoolMgr locking synchronisation
		 */
		public PoolThread(ThreadGroup tg, Object lock, boolean daemonIndicator) {
			super(tg, lock, daemonIndicator);
			setParentAttributes();
			this.start();
		}

		/**
		 * Constructor for PoolThread as DBConnectionMgr.<br>
         * 
		 * @param tg is the ThreadGroup to which this thread belongs.<br>
		 * @param lock is the object that takes care of overall PoolMgr locking/ synchronisation.<br>
		 * @param dbConnector contains the database info required to set the PoolMgr up as a DBConnectionMgr.<br>
		 */

		public PoolThread(ThreadGroup tg, Object lock, boolean daemonIndicator, DBConnector dbConnector) {
			super(tg, lock, daemonIndicator, dbConnector);
			setParentAttributes();
			if (__debug) System.out.println(this.toString()+" constructor about to start");
			this.start();
		}

		protected void setParentAttributes() {
			_nrOfThreadInstances++;
			__threadCounter = _nrOfThreadInstances;
			__debug = _debug;
			__maxIdleTime = _maxIdleTime;
			__minThreads = _minThreads;
			__maxThreads = _maxThreads;
		}

		protected boolean currentThreadInAvailableThreads(ThreadPoolWorker pw) {
			return _availableThreads.contains(this);
		}

		protected int getAvailableThreads() {
			return _availableThreads.size();
		}

		protected int getNrOfPendingJobs() {
			return _pendingJobs.size();
		}

		protected Object getPendingJob() {
				return (Object) _pendingJobs.pop();
		}

		protected void pushIntoAvailableThreads(ThreadPoolWorker pw) {
			_availableThreads.push(this);
		}

		protected void removeCurrentThreadFromQueue(ThreadPoolWorker pw) {
			removeMe(this);
		}

	};

	public ThreadPoolMgr() {
	}

	public ThreadPoolMgr(java.util.Properties props) throws NumberFormatException, IllegalArgumentException
	{
		// First determine whether there are any properties set at all in the properties object
		if (props == null) return;

		Object o = props.getProperty("debug");
		if (((String) o).equals("true")) _debug = true;

		// Check to see whether this Mgr also needs to do database stuff
		o = props.getProperty("databaseConnector");		
		if (o.toString().equalsIgnoreCase("goods")) {
            o = props.getProperty("service");
			_databaseConnector =  new DBConnector(o.toString(), "goods", ".//dbsettings//");
		}
        // Make sure the database is initialized (i.e. the index structure is available)
        if (! _databaseConnector.isServiceDatabaseInitialized()) {
            this.addJob(new DatabaseIndexJob());
            System.out.println("Finished database initialization!");
        }

		int n;
		// Set the maxIdleTime, the maximum amount of time a thread can be idle, if run-out it has to kill itself (&& if #Thread > minThread)
		o = props.getProperty("maxIdleTime");
		if (o != null) {
			n = java.lang.Integer.parseInt((String) o);
			if (n <= 0) throw new IllegalArgumentException(this.getClass().getName()+".Constructor()=> maxIdleTime must be an integral value greater than 0");
			_maxIdleTime = n;
		}

		// Set the maxThreads
		o = props.getProperty("maxThreads");
		if (o != null) {
			n = java.lang.Integer.parseInt((String) o);
			if (n <= 0) throw new IllegalArgumentException(this.getClass().getName()+".Constructor()=> maxThreads must be an integral value greater than 0");
			_maxThreads = n;
		}

		// Set the minThreads
		o = props.getProperty("minThreads");
		if (o != null) {
			n = java.lang.Integer.parseInt((String) o);
			// make sure the minThreads has a proper value
			if (n < 0) {
				n = 0;
			} else
				if (n > _maxThreads)
					n = _maxThreads;
			// Now that minThreads has been determined the threads can be created assuming minThreads > 0
			_minThreads = n;
			// create the min number of threads in advance of the system actually starting up
			if (_minThreads > 0) {
				// If a thread belongs to the minimal set, it should also be a daemon.
				boolean daemonThread = true;
				for (int i = 1; i <= _minThreads; i++) {
					if (_debug) System.out.println("Creating minimum Thread nr " + (_nrOfThreadInstances));
					//PoolThread pt = new PoolThread(_totalThreadGroup, this);
					createNewThread(_totalThreadGroup, daemonThread, _databaseConnector);
					if (_debug) System.out.println("Size of the jobQueue "+ _pendingJobs.size()+ " [req 0]");
				}
				if (_debug) System.out.println("Size of the queue "+ _availableThreads.size()+ " [req "+ _minThreads+ "]");
			}
		}
	} // public ThreadPoolImpl(java.util.Properties props)

	/**
	 * This method is responsible for taking on 'a piece of work' and handing it to a WorkerThread in the pool.<br>
	 * Basically this is where all the dispatching takes place.<br>
	 */
	synchronized public void addJob(GeneralJob job) {
		PoolThread pt;
		// Start by adding the job to be executed to the pending list
		_pendingJobs.push(job);
		// Check to see whether there is a thread available to give the job to
		if (!_availableThreads.isEmpty()) { // there are threads available
			if (_debug) System.out.println("Using an existing thread...");
			// Get a thread from the stack to give the work to
			synchronized (pt = (PoolThread) _availableThreads.pop()) {
				if (_debug) System.out.println(pt.getName() + "lowered threads in the queue to:"+ _availableThreads.size());
				// wake the thread so that it can start reading from the pending jobs list
				pt.wake(new String("Wake up call for " + pt.getName()));
			}
			return;
		} else { // NO threads are readily available so check to see whether more can be added
			if (_maxThreads == -1 || _totalThreadGroup.activeCount() < _maxThreads) {
				// infinite number of threads are allowed or new additional threads can be added
				if (_debug) System.out.println("Creating additional Thread nr " + (_nrOfThreadInstances));
				// As soon as the thread gets constructed he starts working on the pending queue
				// False indicates it is NOT a daemon thread
				createNewThread(_totalThreadGroup, false, _databaseConnector);
				return;
			} else { // No threads available and no more can be created
				// When one of the busy threads is done, it will check the pending queue and will see to this job.
				if (_debug) System.out.println("Job has been placed in the queue because all threads in the pool are busy.");
			}
		}
	}


	synchronized public void addJob(DatabaseJob job) {
    }

	synchronized public ThreadPoolMgrStats getStats() {
		stats.maxThreads = _maxThreads;
		stats.minThreads = _minThreads;
		stats.maxIdleTime = _maxIdleTime;
		stats.pendingJobs = _pendingJobs.size();
		stats.numThreads = _totalThreadGroup.activeCount();
		stats.jobsInProgress = 0;
		stats.databaseConnector = "empty";

		int len = this._totalThreadGroup.activeCount();
		Thread[] tarray = new Thread[len];
		int __stop = _totalThreadGroup.enumerate(tarray);
		for (int i = 0; i <= len-1; i++) { // len-1 because the array starts counting on 0 NOT 1
			if ((tarray[i].getName().startsWith("Thread-")) && (((ThreadPoolWorker) tarray[i])._idle == false)) {
				System.out.println("\tgetStats()=>Processing thread: " + tarray[i].getName());
				stats.jobsInProgress++;
			}
		}
		return (stats);
	}


	public synchronized void closePool() {
		PoolThread pt = null;
		if ((_availableThreads.size() == _minThreads) && (_pendingJobs.isEmpty()) && (_totalThreadGroup.activeCount() == 0)) {
			// _minThreads && NO pending jobs
			while (!_availableThreads.isEmpty()) {
				pt = (PoolThread) _availableThreads.pop();
				pt = null;
				_nrOfThreadInstances--;
			}
		}
	}

	//*****************************************//
	// !!! Important... !!!
	// All private methods must always be called from a synchronized method!!
	//*****************************************//

	/**
	 * This method is called when a new thread needs to be created
	 * @param tg is the ThreadGroup of which the Thread is to be a part of, null indicates NO database thread
	 */
	protected PoolThread createNewThread(final ThreadGroup tg, final boolean daemonThread, final DBConnector dbConnector) {
		if (dbConnector==null) {
			return new PoolThread(tg, this, daemonThread);
		} else {
			return new PoolThread(tg, this, daemonThread, dbConnector);
		}
	}

	// Called by the thread pool to find the number of idle threads
	protected synchronized int findNumberOfIdleThreads() {
		return (_availableThreads.size());
	}

	// Called by a thread to remove itself from the stack/ vector of available threads
	protected void removeMe(Thread pt) {
		_availableThreads.remove(pt);
		_nrOfThreadInstances--;
		if (_debug) System.out.println("Removed "+ pt.getName()+ " from the pool, left "+ _availableThreads.size()+ "["+ _nrOfThreadInstances+ "]");
		pt = null;
	}

}
